require "defines"

function debug(...)
    --print(...)
end

function make_tiny_aabb(p)
    local d = 0.1
    return {{p.x-d, p.y-d}, {p.x+d, p.y+d}}
end    

function get_logistic_chest_at(surface, p)
    local targets = surface.find_entities_filtered{
        type = "logistic-container",
        area = make_tiny_aabb(p)
    }
    if #targets == 1 then
        return targets[1]
    end
    return nil
end 

function get_assembly_at(surface, p)
    local targets = surface.find_entities_filtered{
        type = "assembling-machine",
        area = make_tiny_aabb(p)
    }
    if #targets == 1 then
        return targets[1]
    end
    return nil
end 

function get_desired_output_count(name)
    return game.item_prototypes[name].stack_size
end

function get_desired_input_count(name)
    local default = game.item_prototypes[name].default_request_amount
    if default > 50 then
        return 50
    end
end

function get_item_product_names(products)
    local item_names = {}
    for _,product in ipairs(products) do
        -- MAGIC NUMBER ... ITEM
        if product.type == 0 then
            table.insert(item_names, product.name)
        end
    end
    return item_names
end

function setup_for_output(inserter, provider, assembly)
    local product_names = get_item_product_names(assembly.recipe.products)
    if #product_names ~= 1 then
        debug("Invalid recipe! Can only handle exactly 1 solid result. Not ", #product_names)
        return
    end
    
    local product_name = product_names[1]
    -- Weird API... maybe a bug?
    inserter.set_filter(product_name, 1)
    
    -- No documentation for this? Wiki is wrong...
    inserter.set_circuit_condition(2,{
        condition = {
            first_signal= {
                type = "item",
                name = product_name
            },
            comparator =  "<",
            constant = get_desired_output_count(product_name)
        }
    })
end

function find_free_request_slot(requester, visited, name)
    for i = 1,8 do
        if not visited[i] then
            local item = requester.get_request_slot(i)
            if item == nil or item.name == name then
                return i, item
            end
        end
    end
    return nil, nil
end

function setup_for_input(inserter, requester, assembly)
    if assembly.recipe == nil then
        debug("No recipe set...")
    end
    
    local slots = {}
    local visited = {}
    
    
    for _,ingredient in ipairs(assembly.recipe.ingredients) do
        if ingredient.type == "item" then
            local slot, item  = find_free_request_slot(requester, 
                                                       visited,
                                                       ingredient.name)
            if slot == nil then
                debug("Can't request everything we need... bailing out.")
                return
            end
            visited[slot] = true
            table.insert(slots,{slot=slot,name=ingredient.name,item=item})
        end
    end
    
    local max_filter_count = 5
    local set_filter = #slots <= max_filter_count
    for i,slot in ipairs(slots) do
        local count = get_desired_input_count(slot.name)
        if slot.item and slot.item.count > count then
            count = slot.item.count
        end
        requester.set_request_slot({name = slot.name,count = count}, slot.slot)
        if set_filter then
            inserter.set_filter(slot.name, i)
        end
    end
end

function on_load()
    if global.auto_inserter_loaded == nil then
        for _,force in pairs(game.forces) do
            force.reset_recipes()
            force.reset_technologies()
            local technologies = force.technologies
            local recipes = force.recipes
            if technologies["logistic-robotics"].researched then
                recipes["auto-inserter"].enabled=true
            end
            if technologies["construction-robotics"].researched then
                recipes["auto-inserter"].enabled=true
            end
        end
    end
    global.auto_inserter_loaded = true
    -- Reset memo
    global.is_requester = {}
end

game.on_init(on_load)
game.on_load(on_load)

function is_logistic_requester(entity)
    if global.is_requester[entity.name] == nil then
        local status, err = pcall(function () return entity.get_request_slot(1) end)
        --debug(entity.name, status, err)
        global.is_requester[entity.name] = status
        return status
    end
    return global.is_requester[entity.name]
end

function place_auto_inserter(entity)
    --debug("placed:", entity.name, "on", entity.surface.name)
    --debug("position:", entity.position.x,entity.position.y)
    --debug("drop_position:", entity.drop_position.x,entity.drop_position.y)
    local delta = {x=math.floor(0.5-entity.position.x+entity.drop_position.x),
                   y=math.floor(0.5-entity.position.y+entity.drop_position.y)}
    --debug("delta:",delta.x, delta.y)
    local fetch_position = {x=entity.position.x - delta.x,
                            y=entity.position.y - delta.y}
    --debug("fetch_position:", fetch_position.x,fetch_position.y)
    
    local drop_chest = get_logistic_chest_at(entity.surface, entity.drop_position)
    if drop_chest then
        local assembly = get_assembly_at(entity.surface, fetch_position)
        if assembly then
            setup_for_output(entity, drop_chest, assembly)
            return
        end
        debug("Invalid output setup - need assembly on fetch side.")
        return
    end

    local fetch_chest = get_logistic_chest_at(entity.surface, fetch_position)
    if fetch_chest and is_logistic_requester(fetch_chest) then
        local assembly = get_assembly_at(entity.surface, entity.drop_position)
        if assembly then
            setup_for_input(entity, fetch_chest, assembly)
        end
        debug("Invalid input setup - need assembly on drop side.")
        return
    end
    debug("Invalid setup - need to fetch from requester and place into assembly"
          .. " or fetch from assembly and place into provider.")    
end

game.on_event(defines.events.on_built_entity, function (event)
    if event.created_entity.name ~= "auto-inserter" then return end
    local status, err = pcall(place_auto_inserter, event.created_entity)
    if not status then 
        debug("on_event:", err)
    end
end)

--game.on_event(defines.events.on_robot_built_entity, on_built_entity)

